"""
This module is deprecated. Its contents have been moved to
longitudinal_tomography.data.data_treatment.
This module exists solely for backwards compatibility and can be removed
without notice.
"""

from ..data.data_treatment import *
from warnings import warn

# warn('longitudinal_tomography.utils.data_treatment has been moved to '
#      'longitudinal_tomography.data.data_treatment')
